import * as ts from 'typescript';
import { ViewEncapsulation } from '@angular/core';
import { CompileMetadataResolver, NgModuleResolver, DirectiveResolver, DirectiveNormalizer, HtmlParser, CompilerConfig, PipeResolver, AotSummaryResolver, DomElementSchemaRegistry, StaticSymbolResolver, StaticSymbolCache, StaticReflector, createOfflineCompileUrlResolver, analyzeNgModules, AotCompiler, StyleCompiler, NgModuleCompiler, TypeScriptEmitter, ViewCompiler, TemplateParser, Lexer, Parser } from '@angular/compiler';
import { MetadataCollector, readConfiguration } from '@angular/compiler-cli';
import { PipeSymbol } from './pipe-symbol';
import { DirectiveSymbol } from './directive-symbol';
import { ModuleSymbol } from './module-symbol';
import { TsCompilerAotCompilerTypeCheckHostAdapter } from '@angular/compiler-cli/src/transformers/compiler_host';
import { InjectableCompiler } from '@angular/compiler/src/injectable_compiler';
import { TypeCheckCompiler } from '@angular/compiler/src/view_compiler/type_check_compiler';
/**
 * Creates a proxy which provides us access to the symbols
 * defined in given context (could be lazy loaded module or the root module).
 *
 * @export
 * @class ProjectSymbols
 */
var ProjectSymbols = /** @class */ (function () {
    /**
     * Creates an instance of ProjectSymbols.
     *
     * @param {ts.Program} program
     * @param {ResourceResolver} resourceResolver
     *
     * @memberOf ProjectSymbols
     */
    function ProjectSymbols(tsconfigPath, resourceResolver, errorReporter) {
        this.tsconfigPath = tsconfigPath;
        this.resourceResolver = resourceResolver;
        this.errorReporter = errorReporter;
        var config = readConfiguration(this.tsconfigPath);
        this.options = config.options;
        this.init(this.options, config.rootNames);
    }
    /**
     * Returns the metadata associated to this module.
     *
     * @returns {ModuleSymbol[]}
     *
     * @memberOf ProjectSymbols
     */
    ProjectSymbols.prototype.getModules = function () {
        var _this = this;
        var resultMap = new Map();
        this.getAnalyzedModules().ngModules.forEach(function (m, s) {
            resultMap.set(m.type.reference, m);
        });
        var result = [];
        resultMap.forEach(function (v) {
            return result.push(new ModuleSymbol(_this.program, v.type.reference, _this.metadataResolver, _this.directiveNormalizer, _this.directiveResolver, _this.pipeResolver, _this.reflector, _this.resourceResolver, _this));
        });
        return result;
    };
    /**
     * Returns all the directives available in the context.
     *
     * @returns {DirectiveSymbol[]}
     *
     * @memberOf ProjectSymbols
     */
    ProjectSymbols.prototype.getDirectives = function () {
        var _this = this;
        return this.extractProgramSymbols()
            .filter(function (symbol) { return _this.metadataResolver.isDirective(symbol); })
            .map(function (symbol) {
            return new DirectiveSymbol(_this.program, symbol, _this.metadataResolver, _this.directiveNormalizer, _this.reflector, _this.resourceResolver, _this);
        });
    };
    /**
     * Returns all the pipes available in this module.
     *
     * @returns {PipeSymbol[]}
     *
     * @memberOf ProjectSymbols
     */
    ProjectSymbols.prototype.getPipes = function () {
        var _this = this;
        return this.extractProgramSymbols()
            .filter(function (v) { return _this.metadataResolver.isPipe(v); })
            .map(function (p) { return new PipeSymbol(_this.program, p, _this.pipeResolver, _this.metadataResolver, _this); });
    };
    /**
     * Returns all the providers available in this module.
     *
     * @returns {ProviderSymbol[]}
     *
     * @memberOf ProjectSymbols
     */
    ProjectSymbols.prototype.getProviders = function () {
        var resultSet = new Map();
        this.getModules().forEach(function (m) {
            m.getProviders().forEach(function (p) { return resultSet.set(p.getMetadata(), p); });
        });
        this.getDirectives().forEach(function (d) {
            d.getProviders().forEach(function (p) { return resultSet.set(p.getMetadata(), p); });
            d.getViewProviders().forEach(function (p) { return resultSet.set(p.getMetadata(), p); });
        });
        var finalResult = [];
        resultSet.forEach(function (v) { return finalResult.push(v); });
        return finalResult;
    };
    /**
     * Returns directive based on `ClassDeclaration` node and a filename.
     *
     * @param {ts.ClassDeclaration} declaration
     * @param {string} fileName
     *
     * @memberOf DirectiveSymbol
     */
    ProjectSymbols.prototype.getDirectiveFromNode = function (declaration, fileName) {
        var sourceFile = this.program.getSourceFile(fileName);
        if (!sourceFile) {
            throw new Error("Cannot get the program's source file");
        }
        var identifier = declaration.name;
        if (identifier) {
            return new DirectiveSymbol(this.program, this.reflector.getStaticSymbol(sourceFile.fileName, identifier.text), this.metadataResolver, this.directiveNormalizer, this.reflector, this.resourceResolver, this);
        }
        else {
            return null;
        }
    };
    /** @internal */
    ProjectSymbols.prototype.getAnalyzedModules = function () {
        var analyzedModules = this.analyzedModules;
        if (!analyzedModules) {
            var analyzeHost = {
                isSourceFile: function (filePath) {
                    return true;
                }
            };
            analyzedModules = this.analyzedModules = analyzeNgModules(this.program.getRootFileNames(), analyzeHost, this.staticSymbolResolver, this.metadataResolver);
        }
        return analyzedModules;
    };
    ProjectSymbols.prototype.extractProgramSymbols = function () {
        var _this = this;
        return [].concat.apply([], this.program
            .getSourceFiles()
            .filter(function (f) { return _this.staticSymbolResolver.hasDecorators(f.fileName); })
            .map(function (f) { return _this.staticSymbolResolver.getSymbolsOf(f.fileName); }));
    };
    // private clearCaches() {
    //   this.metadataResolver.clearCache();
    //   this.directiveNormalizer.clearCache();
    // }
    ProjectSymbols.prototype.init = function (options, rootNames) {
        var _this = this;
        this.compilerHost = ts.createCompilerHost(this.options, true);
        // Replace all `\` with a forward slash to align with typescript's `normalizePath`.
        // On Windows, different slashes cause errors while trying to compare module symbols
        rootNames = rootNames.map(function (rootName) { return rootName.replace(/\\/g, '/'); });
        this.staticResolverHost = new TsCompilerAotCompilerTypeCheckHostAdapter(rootNames, this.options, this.compilerHost, new MetadataCollector(), {
            generateFile: function (genFileName, baseFileName) {
                return _this.compiler.emitBasicStub(genFileName, baseFileName);
            },
            findGeneratedFileNames: function (fileName) {
                return _this.compiler.findGeneratedFileNames(fileName);
            }
        });
        this.urlResolver = createOfflineCompileUrlResolver();
        var symbolCache = new StaticSymbolCache();
        this.summaryResolver = new AotSummaryResolver(this.staticResolverHost, symbolCache);
        this.staticSymbolResolver = new StaticSymbolResolver(this.staticResolverHost, symbolCache, this.summaryResolver);
        this.reflector = new StaticReflector(this.summaryResolver, this.staticSymbolResolver, [], []);
        var htmlParser = new HtmlParser();
        var config = new CompilerConfig({
            defaultEncapsulation: ViewEncapsulation.Emulated,
            useJit: false
        });
        this.pipeResolver = new PipeResolver(this.reflector);
        this.directiveResolver = new DirectiveResolver(this.reflector);
        this.directiveNormalizer = new DirectiveNormalizer({
            get: function (url) {
                return _this.staticResolverHost.loadResource(url);
            }
        }, this.urlResolver, htmlParser, config);
        var registry = new DomElementSchemaRegistry();
        this.metadataResolver = new CompileMetadataResolver(config, htmlParser, new NgModuleResolver(this.reflector), this.directiveResolver, this.pipeResolver, this.summaryResolver, registry, this.directiveNormalizer, console, symbolCache, this.reflector);
        var viewCompiler = new ViewCompiler(this.reflector);
        var typeCheckCompiler = new TypeCheckCompiler(options, this.reflector);
        var expressionParser = new Parser(new Lexer());
        var tmplParser = new TemplateParser(config, this.reflector, expressionParser, registry, htmlParser, console, []);
        this.compiler = new AotCompiler(config, options, this.staticResolverHost, this.reflector, this.metadataResolver, tmplParser, new StyleCompiler(this.urlResolver), viewCompiler, typeCheckCompiler, new NgModuleCompiler(this.reflector), new InjectableCompiler(this.reflector, !!options.enableIvy), new TypeScriptEmitter(), this.summaryResolver, this.staticSymbolResolver);
        this.program = ts.createProgram({ rootNames: rootNames, options: options, host: this.staticResolverHost });
    };
    return ProjectSymbols;
}());
export { ProjectSymbols };
//# sourceMappingURL=project-symbols.js.map